@@ -16,7 +16,7 @@ _FP32_OPS: Set[str] = {
     "sin", "cos", "tan",
     # softmax & reductions that accumulate / are variance-based
     "softmax",
-    "sum", "mean", "std",
+    "sum", "mean", "std", "var",
     # losses (compute entirely in fp32)
     "mse", "rmse", "mae", "binarycrossentropy", "categoricalcrossentropy",
     # casting decisions shouldn’t be overridden by autocast
@@ -72,7 +72,8 @@ def should_cast_to_fp16(op_name: str) -> bool:
     return True
 
 
-def maybe_cast_tensor(tensor, target_dtype=None, op_name: str = "unknown") -> 'Tensor':
+def maybe_cast_tensor(tensor, target_dtype=None, op_name: str = "unknown",
+                      memsave: bool = False) -> 'Tensor':
     """
     Cast tensor to appropriate dtype based on autocast context and operation type.
     
@@ -97,13 +98,13 @@ def maybe_cast_tensor(tensor, target_dtype=None, op_name: str = "unknown") -> 'T
         if should_cast_to_fp16(op_name):
             target_dtype = autocast.get_autocast_dtype()
         else:
-            return tensor.cast(ng.float32)
+            return tensor.cast(ng.float32, memsave=memsave)
     
     # Only cast if different from current dtype
-    if tensor.data.dtype == target_dtype:
+    if tensor.data.dtype == target_dtype and not memsave:
         return tensor
-        
-    return tensor.cast(target_dtype)
+
+    return tensor.cast(target_dtype, memsave=memsave)
 
 
 def get_fp32_ops() -> Set[str]:
