@@ -98,7 +98,8 @@ class EinSum(Function, Module):
         self.subscripts = subscripts.replace(" ", "")
     def forward(self, *operands: xp.ndarray) -> xp.ndarray:
         self.operand_shapes = [op.shape for op in operands]
-        return xp.einsum(self.subscripts, *operands)
+        self.input_operands = operands
+        return xp.einsum(self.subscripts, *operands)   
     def backward(self, grad_output: xp.ndarray):
         grads = []
         if '->' in self.subscripts:
@@ -106,7 +107,10 @@ class EinSum(Function, Module):
             input_specs = inputs_str.split(',')
         else:
             input_specs = self.subscripts.split(',')
-            output_str = ''.join(sorted(set(''.join(input_specs))))
+            all_indices = ''.join(input_specs)
+            output_str = ''.join(sorted(set(all_indices) - 
+                                      set([idx for idx in all_indices 
+                                           if all_indices.count(idx) > 1])))
         for i, tensor in enumerate(self.parent_tensors):
             if not tensor.requires_grad:
                 grads.append(None)
@@ -118,10 +122,23 @@ class EinSum(Function, Module):
                     grad_operands.append(operand.data)
                     grad_specs.append(input_specs[j])
             grad_equation = ','.join(grad_specs) + '->' + input_specs[i]
-            grad = xp.einsum(grad_equation, *grad_operands)
-            grad = self._handle_broadcasting(grad, self.operand_shapes[i])
-            grads.append(grad)
+            try:
+                grad = xp.einsum(grad_equation, *grad_operands)
+                grad = self._handle_broadcasting(grad, self.operand_shapes[i])
+                grads.append(grad)
+            except Exception as e:
+                grads.append(xp.zeros_like(self.input_operands[i]))
         return tuple(grads)
+    def _handle_broadcasting(self, grad, original_shape):
+        if grad.shape == original_shape:
+            return grad
+        ndims_added = grad.ndim - len(original_shape)
+        for _ in range(ndims_added):
+            grad = grad.sum(axis=0)
+        for i, (grad_dim, orig_dim) in enumerate(zip(grad.shape, original_shape)):
+            if orig_dim == 1 and grad_dim > 1:
+                grad = grad.sum(axis=i, keepdims=True)
+        return grad
 
 class Transpose(Function, Module):
     name = "Transpose"
