@@ -1,4 +1,5 @@
 from neurograd import xp
+import builtins
 from .base import Function
 from neurograd.nn.module import Module
 
@@ -9,27 +10,22 @@ class Sum(Function, Module):
         Module.__init__(self)
         self.axis = axis
         self.keepdims = keepdims
-    
     def forward(self, x: xp.ndarray) -> xp.ndarray:
         return xp.sum(x, axis=self.axis, keepdims=self.keepdims)
-    
     def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
         x = self.parent_tensors[0]
         if not x.requires_grad:
             return None
-        
         # For Sum, we just need to broadcast grad_output to input shape
-        # Since forward stored keepdims=True result, we can use that for broadcasting
-        grad = self._cast_like(grad_output, x.data)
+        grad = grad_output
         if self.axis is not None and not self.keepdims:
             # Add dimensions back by expanding to match what keepdims=True would give
             for ax in sorted(self.axis if isinstance(self.axis, tuple) else (self.axis,)):
                 # Normalize negative axes
-                ax_norm = ax if ax >= 0 else len(x.data.shape) + ax
+                ax_norm = ax if ax >= 0 else len(x.shape) + ax
                 grad = xp.expand_dims(grad, axis=ax_norm)
-        
         # Broadcast to original shape
-        grad = xp.broadcast_to(grad, x.data.shape)
+        grad = xp.broadcast_to(grad, x.shape)
         return grad
 
 class Mean(Function, Module):
@@ -39,38 +35,30 @@ class Mean(Function, Module):
         Module.__init__(self)
         self.axis = axis
         self.keepdims = keepdims
-    
     def forward(self, x: xp.ndarray) -> xp.ndarray:
         return xp.mean(x, axis=self.axis, keepdims=self.keepdims)
-    
     def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
         x = self.parent_tensors[0]
         if not x.requires_grad:
             return None
-        
         # Calculate number of elements being averaged
         if self.axis is None:
-            n = x.data.size
+            n = x.size
         else:
             if isinstance(self.axis, int):
-                n = x.data.shape[self.axis]
+                n = x.shape[self.axis]
             else:
-                # Handle tuple of axes
                 n = 1
                 for ax in self.axis:
-                    n *= x.data.shape[ax]
-        
+                    n *= x.shape[ax]
+        # In-place division to avoid intermediate array
+        grad_output /= n
         # Expand and broadcast gradient
-        grad = self._cast_like(grad_output, x.data) / n
         if self.axis is not None and not self.keepdims:
-            # Add dimensions back by expanding to match what keepdims=True would give
             for ax in sorted(self.axis if isinstance(self.axis, tuple) else (self.axis,)):
-                # Normalize negative axes
-                ax_norm = ax if ax >= 0 else len(x.data.shape) + ax
-                grad = xp.expand_dims(grad, axis=ax_norm)
-        
-        grad = xp.broadcast_to(grad, x.data.shape)
-        return grad
+                ax_norm = ax if ax >= 0 else len(x.shape) + ax
+                grad_output = xp.expand_dims(grad_output, axis=ax_norm)
+        return xp.broadcast_to(grad_output, x.shape)
 
 class Max(Function, Module):
     name = "Max"
@@ -79,54 +67,39 @@ class Max(Function, Module):
         Module.__init__(self)
         self.axis = axis
         self.keepdims = keepdims
-    
     def forward(self, x: xp.ndarray) -> xp.ndarray:
+        # Cache with keepdims=True to avoid recomputation in backward
         self.max_vals = xp.max(x, axis=self.axis, keepdims=True)
-        return xp.max(x, axis=self.axis, keepdims=self.keepdims)
-    
+        if self.keepdims:
+            return self.max_vals
+        else:
+            return xp.squeeze(self.max_vals, axis=self.axis)
     def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
         x = self.parent_tensors[0]
         if not x.requires_grad:
             return None
-        
-        # Create mask for maximum values
-        mask = (x.data == self.max_vals).astype(x.data.dtype)
-        
-        # Handle ties by distributing gradient equally
+        # Create gradient mask for max elements, splitting ties evenly along the axis.
+        # Use self.max_vals which already has keepdims=True for broadcasting
+        mask = (x.data == self.max_vals).astype(x.dtype)
         if self.axis is None:
-            # Global max - distribute among all max elements
-            mask = mask / xp.sum(mask)
+            mask /= xp.sum(mask)  # In-place division
         else:
-            # Axis-wise max - distribute among max elements along each reduced dimension
-            if isinstance(self.axis, int):
-                count = xp.sum(mask, axis=self.axis, keepdims=True)
-            else:
-                count = xp.sum(mask, axis=self.axis, keepdims=True)
-            # Avoid division by zero
+            count = xp.sum(mask, axis=self.axis, keepdims=True)
             count = xp.where(count == 0, 1, count)
-            mask = mask / count
-        
+            mask /= count  # In-place division
         # Expand and broadcast gradient
-        grad = self._cast_like(grad_output, x.data)
+        grad = grad_output
         if self.axis is not None and not self.keepdims:
-            # We need to expand grad_output to have the same number of dimensions as x.data
-            # For each reduced axis, we need to add a dimension of size 1
-            
-            # Convert negative axes to positive
-            ndim = len(x.data.shape)
+            ndim = x.ndim
             if isinstance(self.axis, int):
                 axis_list = [self.axis if self.axis >= 0 else ndim + self.axis]
             else:
                 axis_list = [ax if ax >= 0 else ndim + ax for ax in self.axis]
-            
-            # Sort axes for consistent expansion
             axis_list = sorted(axis_list)
-            
-            # Add dimensions one by one
             for ax in axis_list:
                 grad = xp.expand_dims(grad, axis=ax)
-        
-        grad = xp.broadcast_to(grad, x.data.shape) * mask
+        grad = xp.broadcast_to(grad, x.shape)
+        grad *= mask  # In-place multiplication
         return grad
 
 class Min(Function, Module):
@@ -136,119 +109,134 @@ class Min(Function, Module):
         Module.__init__(self)
         self.axis = axis
         self.keepdims = keepdims
-    
     def forward(self, x: xp.ndarray) -> xp.ndarray:
+        # Cache with keepdims=True to avoid recomputation in backward
         self.min_vals = xp.min(x, axis=self.axis, keepdims=True)
-        return xp.min(x, axis=self.axis, keepdims=self.keepdims)
-    
+        if self.keepdims:
+            return self.min_vals
+        else:
+            return xp.squeeze(self.min_vals, axis=self.axis)
     def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
         x = self.parent_tensors[0]
         if not x.requires_grad:
             return None
-        
-        # Create mask for minimum values
-        mask = (x.data == self.min_vals).astype(x.data.dtype)
-        
-        # Handle ties by distributing gradient equally
+        # Create gradient mask for min elements, splitting ties evenly along the axis.
+        # Use self.min_vals which already has keepdims=True for broadcasting  
+        mask = (x.data == self.min_vals).astype(x.dtype)
         if self.axis is None:
-            # Global min - distribute among all min elements
-            mask = mask / xp.sum(mask)
+            mask /= xp.sum(mask)  # In-place division
         else:
-            # Axis-wise min - distribute among min elements along each reduced dimension
-            if isinstance(self.axis, int):
-                count = xp.sum(mask, axis=self.axis, keepdims=True)
-            else:
-                count = xp.sum(mask, axis=self.axis, keepdims=True)
-            # Avoid division by zero
+            count = xp.sum(mask, axis=self.axis, keepdims=True)
             count = xp.where(count == 0, 1, count)
-            mask = mask / count
-        
+            mask /= count  # In-place division
         # Expand and broadcast gradient
-        grad = self._cast_like(grad_output, x.data)
+        grad = grad_output
         if self.axis is not None and not self.keepdims:
-            # We need to expand grad_output to have the same number of dimensions as x.data
-            # For each reduced axis, we need to add a dimension of size 1
-            
-            # Convert negative axes to positive
-            ndim = len(x.data.shape)
+            ndim = x.ndim
             if isinstance(self.axis, int):
                 axis_list = [self.axis if self.axis >= 0 else ndim + self.axis]
             else:
                 axis_list = [ax if ax >= 0 else ndim + ax for ax in self.axis]
-            
-            # Sort axes for consistent expansion
             axis_list = sorted(axis_list)
-            
-            # Add dimensions one by one
             for ax in axis_list:
                 grad = xp.expand_dims(grad, axis=ax)
-        
-        grad = xp.broadcast_to(grad, x.data.shape) * mask
+        grad = xp.broadcast_to(grad, x.shape)
+        grad *= mask  # In-place multiplication
         return grad
 
-
 class Std(Function, Module):
     name = "Std"
-    def __init__(self, axis=None, keepdims=False, ddof=0):
+    def __init__(self, axis=None, keepdims=False, eps=1e-8):
         Function.__init__(self)
         Module.__init__(self)
         self.axis = axis
         self.keepdims = keepdims
-        self.ddof = ddof  # Delta degrees of freedom
-    
+        self.eps = eps
     def forward(self, x: xp.ndarray) -> xp.ndarray:
-        return xp.std(x, axis=self.axis, keepdims=self.keepdims, ddof=self.ddof)
-    
+        # Cache mean and std for backward pass to avoid recomputation
+        self.mean = xp.mean(x, axis=self.axis, keepdims=True)
+        self.std_vals = xp.std(x, axis=self.axis, keepdims=True)
+        if self.keepdims:
+            return self.std_vals
+        else:
+            return xp.squeeze(self.std_vals, axis=self.axis)
     def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
         x = self.parent_tensors[0]
         if not x.requires_grad:
             return None
-        
-        # Calculate mean and variance for gradient computation
-        mean_val = xp.mean(x.data, axis=self.axis, keepdims=True)
-        var_val = xp.var(x.data, axis=self.axis, keepdims=True, ddof=self.ddof)
-        std_val = xp.sqrt(var_val)
-        
-        # Calculate number of elements
+        # Use cached values to avoid recomputation
+        mean = self.mean
+        std = self.std_vals
+        # Count elements along the reduced axes (population)
         if self.axis is None:
-            n = x.data.size - self.ddof
+            n = x.size
         else:
-            if isinstance(self.axis, int):
-                n = x.data.shape[self.axis] - self.ddof
-            else:
-                n = 1
-                for ax in self.axis:
-                    n *= x.data.shape[ax]
-                n -= self.ddof
-        
-        # Avoid division by zero
-        std_val = xp.where(std_val == 0, 1e-8, std_val)
-        
-        # Gradient: d/dx std(x) = (x - mean) / (n * std)
-        grad = (x.data - mean_val) / (n * std_val)
-        
-        # Expand grad_output to match input shape
-        grad_out = self._cast_like(grad_output, x.data)
+            axes = (self.axis,) if isinstance(self.axis, int) else tuple(self.axis)
+            axes = tuple(a if a >= 0 else x.ndim + a for a in axes)
+            n = 1
+            for a in axes:
+                n *= x.shape[a]
+        # Avoid divide-by-zero
+        std_safe = xp.where(std == 0, self.eps, std)
+        # d/dx std(x) = (x - mean) / (n * std)   (population)
+        # Combine operations to reduce intermediates
+        base_grad = (x.data - mean) / (n * std_safe)
+        # If keepdims=False, expand grad_output back along reduced axes
+        go = grad_output
         if self.axis is not None and not self.keepdims:
-            # Normalize negative axes to positive indices
-            ndim = len(x.data.shape)
-            if isinstance(self.axis, int):
-                axis_normalized = self.axis if self.axis >= 0 else ndim + self.axis
-                grad_out = xp.expand_dims(grad_out, axis=axis_normalized)
-            else:
-                # Sort axes in reverse order for proper expansion
-                axes_normalized = []
-                for ax in self.axis:
-                    axes_normalized.append(ax if ax >= 0 else ndim + ax)
-                for ax in sorted(axes_normalized, reverse=True):
-                    grad_out = xp.expand_dims(grad_out, axis=ax)
-        
-        # Broadcast and multiply
-        grad_out = xp.broadcast_to(grad_out, x.data.shape)
-        grad = grad * grad_out
-        
-        return grad
+            axes = (self.axis,) if isinstance(self.axis, int) else tuple(self.axis)
+            axes = tuple(a if a >= 0 else x.ndim + a for a in axes)
+            for a in sorted(axes):
+                go = xp.expand_dims(go, axis=a)
+        # Broadcast and apply chain rule
+        go = xp.broadcast_to(go, x.shape)
+        base_grad *= go  # In-place multiplication
+        return base_grad
 
+class Var(Function, Module):
+    name = "Var"
+    def __init__(self, axis=None, keepdims=False, ddof=0, eps=1e-8):
+        Function.__init__(self)
+        Module.__init__(self)
+        self.axis = axis
+        self.keepdims = keepdims
+        self.ddof = ddof
+        self.eps = eps
+    def forward(self, x: xp.ndarray) -> xp.ndarray:
+        # Cache mean for backward pass to avoid recomputation
+        self.mean = xp.mean(x, axis=self.axis, keepdims=True)
+        var_vals = xp.var(x, axis=self.axis, keepdims=self.keepdims, ddof=self.ddof)
+        return var_vals
+    def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
+        x = self.parent_tensors[0]
+        if not x.requires_grad:
+            return None
+        # Use cached mean to avoid recomputation
+        mean = self.mean
+        # Count elements along the reduced axes
+        if self.axis is None:
+            n = x.size
+            axes = tuple(range(x.ndim))
+        else:
+            axes = (self.axis,) if isinstance(self.axis, int) else tuple(self.axis)
+            axes = tuple(a if a >= 0 else x.ndim + a for a in axes)
+            n = 1
+            for a in axes:
+                n *= x.shape[a]
+        # Denominator n - ddof, avoid divide-by-zero or negative
+        denom = builtins.max(n - self.ddof, self.eps)
+        # d/dx var(x) = 2(x - mean)/(n - ddof)
+        # Combine operations to reduce intermediates
+        base_grad = 2.0 * (x.data - mean) / denom
+        # If keepdims=False, expand grad_output back along reduced axes
+        go = grad_output
+        if self.axis is not None and not self.keepdims:
+            for a in sorted(axes):
+                go = xp.expand_dims(go, axis=a)
+        # Broadcast and apply chain rule
+        go = xp.broadcast_to(go, x.shape)
+        base_grad *= go  # In-place multiplication
+        return base_grad
 
 def sum(x, axis=None, keepdims=False):
     return Sum(axis=axis, keepdims=keepdims)(x)
@@ -258,5 +246,7 @@ def max(x, axis=None, keepdims=False):
     return Max(axis=axis, keepdims=keepdims)(x)
 def min(x, axis=None, keepdims=False):
     return Min(axis=axis, keepdims=keepdims)(x)
-def std(x, axis=None, keepdims=False, ddof=0):
-    return Std(axis=axis, keepdims=keepdims, ddof=ddof)(x)
+def std(x, axis=None, keepdims=False):
+    return Std(axis=axis, keepdims=keepdims)(x)
+def var(x, axis=None, keepdims=False):
+    return Var(axis=axis, keepdims=keepdims)(x)
\ No newline at end of file
